
-----------------------------------------------------------------------------------------
--
-- Copyright (c) 2009-2012 ProactiveRA
--
-- This script is free software; you can redistribute it and/or modify it under
-- the terms of the GNU Lesser General Public License as published by the Free
-- Software Foundation; either version 3 of the License, or (at your option)
-- any later version.
--
-- This script is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-- FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
-- details.
--
-- You should have received a copy of the GNU Lesser General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-----------------------------------------------------------------------------------------



DROP FUNCTION F_KPI_FORMULA_EVAL;

CREATE OR REPLACE FUNCTION F_KPI_FORMULA_EVAL (
	V_FORMULA IN VARCHAR2,
	V_PARAM_A IN NUMBER,
	V_PARAM_B IN NUMBER DEFAULT NULL,
	V_PARAM_C IN NUMBER DEFAULT NULL,
	V_DATE IN DATE DEFAULT TRUNC(sysdate-1, 'DD')
  )
RETURN TYPE_KPI_RESULT_ARRAY IS
  V_COUNT NUMBER(15,4);
  V_COUNT_A NUMBER(15,4);
  V_COUNT_B NUMBER(15,4);
  V_COUNT_C NUMBER(15,4);
  V_FREQUENCY NUMBER(1);
  BREAK     NUMBER(1);
  COMMAND_LINE VARCHAR2(900);
  V_FORMULA_STRING VARCHAR2(900);
  TYPE T_DYN_CURSOR IS REF CURSOR;
  C1 T_DYN_CURSOR;
  KPI_RESULT_ARRAY TYPE_KPI_RESULT_ARRAY:=TYPE_KPI_RESULT_ARRAY();
BEGIN
 BREAK:=0;

 SELECT MAX(SQL_DYNA_PARAM_TYPE)
 INTO V_FREQUENCY
 FROM REF_KPI_DEFINITION
 WHERE KPI_ID IN (V_PARAM_A, V_PARAM_B, V_PARAM_C);

 BEGIN
  SELECT NVL(CALCULATED_VALUE, 0) INTO V_COUNT_A
  FROM KPI_CALCULATED_VALUES_HISTORY
  WHERE KPI_ID=V_PARAM_A
  AND CALCULATED_DATE=
	(SELECT MAX(CALCULATED_DATE)
 	 FROM KPI_CALCULATED_VALUES_HISTORY
	 WHERE KPI_ID=V_PARAM_A
	 AND (
            (
	          V_FREQUENCY = 4
	      AND TRUNC(CALCULATED_DATE, 'DD') >= V_DATE
	      AND TRUNC(CALCULATED_DATE, 'DD') < (V_DATE+7)
            )
            OR
	    (
	          V_FREQUENCY = 5
	      AND TRUNC(CALCULATED_DATE, 'DD') >= TRUNC(V_DATE, 'MM')
	      AND TRUNC(CALCULATED_DATE, 'DD') < ADD_MONTHS(TRUNC(V_DATE, 'MM'), +1)
	    )
          ));
 EXCEPTION
  WHEN NO_DATA_FOUND THEN  
     BEGIN
	 -- NOTHING FOUND FOR THAT PERIOD, USE LAST KNOWN CALCULATED VALUE
	  SELECT NVL(CALCULATED_VALUE, 0) INTO V_COUNT_A
	  FROM KPI_CALCULATED_VALUES_HISTORY
	  WHERE KPI_ID=V_PARAM_A
	  AND CALCULATED_DATE=
		(SELECT MAX(CALCULATED_DATE)
	 	 FROM KPI_CALCULATED_VALUES_HISTORY
		 WHERE KPI_ID=V_PARAM_A);
	 EXCEPTION
	  WHEN NO_DATA_FOUND THEN  
	     V_COUNT:=0;
	     BREAK:=1;
     END;
 END;

  COMMAND_LINE:='REPLACE(''' || V_FORMULA || ''', ''A'', ''' || TO_CHAR(V_COUNT_A) || ''')';

  IF ((V_PARAM_B IS NOT NULL) AND (BREAK=0)) THEN
     BEGIN
	SELECT NVL(CALCULATED_VALUE, 0) INTO V_COUNT_B
	FROM KPI_CALCULATED_VALUES_HISTORY
	WHERE KPI_ID=V_PARAM_B
	AND CALCULATED_DATE=
		(SELECT MAX(CALCULATED_DATE)
	 	 FROM KPI_CALCULATED_VALUES_HISTORY
		 WHERE KPI_ID=V_PARAM_B
		 AND (
	            (
		          V_FREQUENCY = 4
		      AND TRUNC(CALCULATED_DATE, 'DD') >= V_DATE
		      AND TRUNC(CALCULATED_DATE, 'DD') < (V_DATE+7)
	            )
	            OR
		    (
		          V_FREQUENCY = 5
		      AND TRUNC(CALCULATED_DATE, 'DD') >= TRUNC(V_DATE, 'MM')
		      AND TRUNC(CALCULATED_DATE, 'DD') < ADD_MONTHS(TRUNC(V_DATE, 'MM'), +1)
		    )
	          ));
     EXCEPTION
       WHEN NO_DATA_FOUND THEN  
	BEGIN
		 -- NOTHING FOUND FOR THAT PERIOD, USE LAST KNOWN CALCULATED VALUE
		SELECT NVL(CALCULATED_VALUE, 0) INTO V_COUNT_B
		FROM KPI_CALCULATED_VALUES_HISTORY
		WHERE KPI_ID=V_PARAM_B
		AND CALCULATED_DATE=
			(SELECT MAX(CALCULATED_DATE)
		 	 FROM KPI_CALCULATED_VALUES_HISTORY
			 WHERE KPI_ID=V_PARAM_B);
		 EXCEPTION
		  WHEN NO_DATA_FOUND THEN  
		     V_COUNT:=0;
		     BREAK:=1;
	END;
     END;

	COMMAND_LINE:='REPLACE(' || COMMAND_LINE || ', ''B'', ''' || TO_CHAR(V_COUNT_B) || ''')';

  END IF;

  IF ((V_PARAM_C IS NOT NULL) AND (BREAK=0)) THEN
      BEGIN
	SELECT NVL(CALCULATED_VALUE, 0) INTO V_COUNT_C
	FROM KPI_CALCULATED_VALUES_HISTORY
	WHERE KPI_ID=V_PARAM_C
	AND CALCULATED_DATE=
		(SELECT MAX(CALCULATED_DATE)
	 	 FROM KPI_CALCULATED_VALUES_HISTORY
		 WHERE KPI_ID=V_PARAM_C
		 AND (
	            (
		          V_FREQUENCY = 4
		      AND TRUNC(CALCULATED_DATE, 'DD') >= V_DATE
		      AND TRUNC(CALCULATED_DATE, 'DD') < (V_DATE+7)
	            )
	            OR
		    (
		          V_FREQUENCY = 5
		      AND TRUNC(CALCULATED_DATE, 'DD') >= TRUNC(V_DATE, 'MM')
		      AND TRUNC(CALCULATED_DATE, 'DD') < ADD_MONTHS(TRUNC(V_DATE, 'MM'), +1)
		    )
	          ));
     EXCEPTION
       WHEN NO_DATA_FOUND THEN  
	BEGIN
		--  NOTHING FOUND FOR THAT PERIOD, USE LAST KNOWN CALCULATED VALUE
		SELECT NVL(CALCULATED_VALUE, 0) INTO V_COUNT_C
		FROM KPI_CALCULATED_VALUES_HISTORY
		WHERE KPI_ID=V_PARAM_C
		AND CALCULATED_DATE=
			(SELECT MAX(CALCULATED_DATE)
		 	 FROM KPI_CALCULATED_VALUES_HISTORY
			 WHERE KPI_ID=V_PARAM_C);
		 EXCEPTION
		  WHEN NO_DATA_FOUND THEN  
		     V_COUNT:=0;
		     BREAK:=1;
	END;
     END;

	COMMAND_LINE:='REPLACE(' || COMMAND_LINE || ', ''C'', ''' || TO_CHAR(V_COUNT_C) || ''')';

  END IF;

  COMMAND_LINE:='SELECT ' || COMMAND_LINE || ' FROM DUAL';

  IF (BREAK=0) THEN
  OPEN C1 FOR COMMAND_LINE;
  LOOP
      FETCH C1 INTO V_FORMULA_STRING;
      EXIT WHEN C1%NOTFOUND;
  END LOOP;
  CLOSE C1;

BEGIN
  COMMAND_LINE:='SELECT ' || V_FORMULA_STRING || ' FROM DUAL';
  OPEN C1 FOR COMMAND_LINE;
  LOOP
      FETCH C1 INTO V_COUNT;
      EXIT WHEN C1%NOTFOUND;
  END LOOP;
  CLOSE C1;
EXCEPTION
    WHEN OTHERS THEN
	V_COUNT:=0;
END;

  END IF;
 
  KPI_RESULT_ARRAY.EXTEND;
  KPI_RESULT_ARRAY(1):=V_COUNT;

  RETURN KPI_RESULT_ARRAY;

END F_KPI_FORMULA_EVAL;
/

