
-----------------------------------------------------------------------------------------
--
-- Copyright (c) 2009-2012 ProactiveRA
--
-- This script is free software; you can redistribute it and/or modify it under
-- the terms of the GNU Lesser General Public License as published by the Free
-- Software Foundation; either version 3 of the License, or (at your option)
-- any later version.
--
-- This script is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-- FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
-- details.
--
-- You should have received a copy of the GNU Lesser General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-----------------------------------------------------------------------------------------



DROP PROCEDURE P_KPI_EXECUTE;

CREATE OR REPLACE PROCEDURE P_KPI_EXECUTE(
	V_KPI_ID IN NUMBER, 
	V_KPI_PASSED_PARAM VARCHAR2 DEFAULT NULL)
IS
  COMMAND_LINE VARCHAR2(3000);
  V_FUNCTION	 VARCHAR2(60);
  V_PARAM	 VARCHAR2(300);
  V_DYNA_PARAM_NAME    VARCHAR2(35);
  V_DYNA_PARAM_TYPE    NUMBER(1); -- 1=Integer, 2=Date, 3=Platform-Load#, 4=Week, 5=Month
  V_DYNA_PARAM_LASTVAL VARCHAR2(35); 
  V_DYNA_PARAM_FORMAT  VARCHAR2(20); 
  V_NEW_CALC_VALUE NUMBER(15,4);
  V_OLD_CALC_VALUE NUMBER(15,4);
  V_WINDOW NUMBER(3);
  V_NEW_DYNA_PARAM_VALUE VARCHAR2(35);
  TYPE T_DYN_CURSOR IS REF CURSOR;
  V_LOAD_A NUMBER(3);
  V_LOAD_B NUMBER(3);
  C1 T_DYN_CURSOR;
  KPI_RESULT_ARRAY TYPE_KPI_RESULT_ARRAY:=TYPE_KPI_RESULT_ARRAY();
BEGIN

  -- First, get the last value calculated and last parameter used for the KPI execution
  BEGIN
	  SELECT CALCULATED_VALUE, PARAM_USED
	  INTO V_OLD_CALC_VALUE, V_DYNA_PARAM_LASTVAL
	  FROM KPI_CALCULATED_VALUES
	  WHERE KPI_ID=V_KPI_ID;
  EXCEPTION
     WHEN NO_DATA_FOUND THEN  

	-- No row in KPI_CALCULATED_VALUES, insert one

	SELECT SQL_DYNA_PARAM_TYPE, SQL_DYNA_PARAM_FORMAT
	INTO  V_DYNA_PARAM_TYPE, V_DYNA_PARAM_FORMAT
	FROM REF_KPI_DEFINITION
	WHERE KPI_ID=V_KPI_ID;
	
	IF (V_DYNA_PARAM_TYPE = 1) THEN
		INSERT INTO KPI_CALCULATED_VALUES
		(KPI_ID, CALCULATED_DATE, CALCULATED_VALUE, PARAM_USED)
		SELECT KPI_ID, SYSDATE, 0, 0
		FROM REF_KPI_DEFINITION
		WHERE KPI_ID=V_KPI_ID;

		V_DYNA_PARAM_LASTVAL:=0;
	END IF;

	IF (V_DYNA_PARAM_TYPE = 2) THEN
		INSERT INTO KPI_CALCULATED_VALUES
		(KPI_ID, CALCULATED_DATE, CALCULATED_VALUE, PARAM_USED)
		SELECT KPI_ID, SYSDATE, 0, TO_CHAR(sysdate-2, SQL_DYNA_PARAM_FORMAT)
		FROM REF_KPI_DEFINITION
		WHERE KPI_ID=V_KPI_ID;

		V_DYNA_PARAM_LASTVAL:=TO_CHAR(sysdate-2, V_DYNA_PARAM_FORMAT);
	END IF;

	IF (V_DYNA_PARAM_TYPE = 3) THEN
		INSERT INTO KPI_CALCULATED_VALUES
		(KPI_ID, CALCULATED_DATE, CALCULATED_VALUE, PARAM_USED)
		SELECT KPI_ID, SYSDATE, 0, null
		FROM REF_KPI_DEFINITION
		WHERE KPI_ID=V_KPI_ID;

		V_DYNA_PARAM_LASTVAL:=null;
	END IF;

	IF (V_DYNA_PARAM_TYPE = 4) THEN
		INSERT INTO KPI_CALCULATED_VALUES
		(KPI_ID, CALCULATED_DATE, CALCULATED_VALUE, PARAM_USED)
		SELECT KPI_ID, SYSDATE, 0, TO_CHAR(TRUNC(sysdate, 'WW')-7, SQL_DYNA_PARAM_FORMAT)
		FROM REF_KPI_DEFINITION
		WHERE KPI_ID=V_KPI_ID;

		V_DYNA_PARAM_LASTVAL:=TO_CHAR(TRUNC(sysdate, 'WW')-7, V_DYNA_PARAM_FORMAT);
	END IF;

	IF (V_DYNA_PARAM_TYPE = 5) THEN
		INSERT INTO KPI_CALCULATED_VALUES
		(KPI_ID, CALCULATED_DATE, CALCULATED_VALUE, PARAM_USED)
		SELECT KPI_ID, SYSDATE, 0, TO_CHAR(ADD_MONTHS(TRUNC(sysdate, 'MM'), -1), SQL_DYNA_PARAM_FORMAT)
		FROM REF_KPI_DEFINITION
		WHERE KPI_ID=V_KPI_ID;

		V_DYNA_PARAM_LASTVAL:=TO_CHAR(ADD_MONTHS(TRUNC(sysdate, 'MM'), -1), V_DYNA_PARAM_FORMAT);
	END IF;

	V_OLD_CALC_VALUE:=0;
  END;

  SELECT SQL_EXEC_FUNCTION, 
	 SQL_EXEC_PARAM,
	 SQL_DYNA_PARAM_NAME,
	 SQL_DYNA_PARAM_TYPE,
         SQL_DYNA_PARAM_FORMAT,
	 RECALCULATION_WINDOW
  INTO  V_FUNCTION, 
	V_PARAM,
	V_DYNA_PARAM_NAME,
 	V_DYNA_PARAM_TYPE,
        V_DYNA_PARAM_FORMAT,
  	V_WINDOW
  FROM REF_KPI_DEFINITION
  WHERE KPI_ID=V_KPI_ID;

 IF (V_WINDOW > 0) THEN

-- POSITION THE LAST-VALUE BASED ONE THE RECALCULATION WINDOW
  IF (V_DYNA_PARAM_TYPE = 1) THEN
	COMMAND_LINE:='SELECT ';
	COMMAND_LINE:=COMMAND_LINE || V_DYNA_PARAM_LASTVAL || '-' || TO_CHAR(V_WINDOW);
	COMMAND_LINE:=COMMAND_LINE || ' FROM DUAL';
  END IF;

  IF (V_DYNA_PARAM_TYPE = 2) THEN
	COMMAND_LINE:='SELECT ';
	COMMAND_LINE:=COMMAND_LINE || 'TO_CHAR(TO_DATE(''' || V_DYNA_PARAM_LASTVAL || ''', ''';
	COMMAND_LINE:=COMMAND_LINE || V_DYNA_PARAM_FORMAT || ''')' || '-' || TO_CHAR(V_WINDOW) || ', ';
	COMMAND_LINE:=COMMAND_LINE || '''' || V_DYNA_PARAM_FORMAT || ''')';
	COMMAND_LINE:=COMMAND_LINE || ' FROM DUAL';
  END IF;

  IF (V_DYNA_PARAM_TYPE = 4) THEN
	COMMAND_LINE:='SELECT ';
	COMMAND_LINE:=COMMAND_LINE || 'TO_CHAR(TRUNC(SYSDATE, ''WW'')-(7*(1+' || TO_CHAR(V_WINDOW) || ')), ';
	COMMAND_LINE:=COMMAND_LINE || '''' || V_DYNA_PARAM_FORMAT || ''')';
	COMMAND_LINE:=COMMAND_LINE || ' FROM DUAL';
  END IF;

  IF (V_DYNA_PARAM_TYPE = 5) THEN
	COMMAND_LINE:='SELECT ';
	COMMAND_LINE:=COMMAND_LINE || 'TO_CHAR(ADD_MONTHS(TRUNC(SYSDATE, ''MM''),-(1+' || TO_CHAR(V_WINDOW) || ')), ';
	COMMAND_LINE:=COMMAND_LINE || '''' || V_DYNA_PARAM_FORMAT || ''')';
	COMMAND_LINE:=COMMAND_LINE || ' FROM DUAL';
  END IF;

 IF (V_KPI_PASSED_PARAM IS NULL) THEN
  IF (V_DYNA_PARAM_TYPE != 3) THEN
	  OPEN C1 FOR COMMAND_LINE;
		FETCH C1 INTO V_DYNA_PARAM_LASTVAL;
	  CLOSE C1;
  END IF;
 ELSE
   V_NEW_DYNA_PARAM_VALUE:=V_KPI_PASSED_PARAM;
 END IF;

 END IF;

  FOR I IN 0..V_WINDOW LOOP

  -- Calculate the new dynamic parameter value to be used

  IF (V_DYNA_PARAM_TYPE = 1) THEN
	COMMAND_LINE:='SELECT ';
	COMMAND_LINE:=COMMAND_LINE || V_DYNA_PARAM_LASTVAL || '+1';
	COMMAND_LINE:=COMMAND_LINE || ' FROM DUAL';
  END IF; 
  IF (V_DYNA_PARAM_TYPE = 2) THEN
	COMMAND_LINE:='SELECT ';
	COMMAND_LINE:=COMMAND_LINE || 'TO_CHAR(LEAST(TRUNC(SYSDATE, ''DD''), TO_DATE(''' || V_DYNA_PARAM_LASTVAL || ''', ''';
	COMMAND_LINE:=COMMAND_LINE || V_DYNA_PARAM_FORMAT || ''')' || '+1), ';
	COMMAND_LINE:=COMMAND_LINE || '''' || V_DYNA_PARAM_FORMAT || ''')';
	COMMAND_LINE:=COMMAND_LINE || ' FROM DUAL';
  END IF;

  IF (V_DYNA_PARAM_TYPE = 4) THEN
	COMMAND_LINE:='SELECT ';
	COMMAND_LINE:=COMMAND_LINE || 'TO_CHAR(LEAST(TRUNC(SYSDATE, ''WW''), TO_DATE(''' || V_DYNA_PARAM_LASTVAL || ''', ''';
	COMMAND_LINE:=COMMAND_LINE || V_DYNA_PARAM_FORMAT || ''')' || '+7), ';
	COMMAND_LINE:=COMMAND_LINE || '''' || V_DYNA_PARAM_FORMAT || ''')';
	COMMAND_LINE:=COMMAND_LINE || ' FROM DUAL';
  END IF;

  IF (V_DYNA_PARAM_TYPE = 5) THEN
	COMMAND_LINE:='SELECT ';
	COMMAND_LINE:=COMMAND_LINE || 'TO_CHAR(LEAST(TRUNC(SYSDATE, ''MM''), ADD_MONTHS(TO_DATE(''' || V_DYNA_PARAM_LASTVAL || ''', ''';
	COMMAND_LINE:=COMMAND_LINE || V_DYNA_PARAM_FORMAT || '''),' || '1)), ';
	COMMAND_LINE:=COMMAND_LINE || '''' || V_DYNA_PARAM_FORMAT || ''')';
	COMMAND_LINE:=COMMAND_LINE || ' FROM DUAL';
  END IF;


 IF (V_KPI_PASSED_PARAM IS NULL) THEN

  IF (V_DYNA_PARAM_TYPE != 3) THEN
	  OPEN C1 FOR COMMAND_LINE;
		FETCH C1 INTO V_NEW_DYNA_PARAM_VALUE;
	  CLOSE C1;
  END IF;

 ELSE
   V_NEW_DYNA_PARAM_VALUE:=V_KPI_PASSED_PARAM;
 END IF;

  COMMAND_LINE:='SELECT ' || V_FUNCTION || '(' || V_PARAM ; 
  IF (V_DYNA_PARAM_TYPE = 1) THEN
	COMMAND_LINE:=COMMAND_LINE || ', ' || V_NEW_DYNA_PARAM_VALUE;
  END IF; 
  IF (V_DYNA_PARAM_TYPE = 2) THEN
	COMMAND_LINE:=COMMAND_LINE || ', ' || 'TO_DATE(''' || V_NEW_DYNA_PARAM_VALUE || ''', ''' || V_DYNA_PARAM_FORMAT || ''')';
  END IF; 
  IF ((V_DYNA_PARAM_TYPE = 3) AND (V_KPI_PASSED_PARAM IS NOT NULL)) THEN
	COMMAND_LINE:=COMMAND_LINE || ', ' || V_NEW_DYNA_PARAM_VALUE;
  END IF;
  IF (V_DYNA_PARAM_TYPE = 4) THEN
	COMMAND_LINE:=COMMAND_LINE || ', ' || 'TO_DATE(''' || V_NEW_DYNA_PARAM_VALUE || ''', ''' || V_DYNA_PARAM_FORMAT || ''')';
  END IF; 
  IF (V_DYNA_PARAM_TYPE = 5) THEN
	COMMAND_LINE:=COMMAND_LINE || ', ' || 'TO_DATE(''' || V_NEW_DYNA_PARAM_VALUE || ''', ''' || V_DYNA_PARAM_FORMAT || ''')';
  END IF; 
  COMMAND_LINE:=COMMAND_LINE || ') FROM DUAL';

  OPEN C1 FOR COMMAND_LINE;
      FETCH C1 INTO KPI_RESULT_ARRAY;
  CLOSE C1;

  V_NEW_CALC_VALUE:=KPI_RESULT_ARRAY(1);
  IF (V_DYNA_PARAM_TYPE = 3) THEN
	V_LOAD_A:=KPI_RESULT_ARRAY(2);
	V_LOAD_B:=KPI_RESULT_ARRAY(3);
	V_NEW_DYNA_PARAM_VALUE:='A:' || TO_CHAR(V_LOAD_A);
	IF (V_LOAD_B > 0) THEN
		V_NEW_DYNA_PARAM_VALUE:=V_NEW_DYNA_PARAM_VALUE || ', B:' || TO_CHAR(V_LOAD_B);
	END IF;
  END IF;

  UPDATE KPI_CALCULATED_VALUES
  SET    CALCULATED_DATE=SYSDATE,
	 CALCULATED_VALUE=V_NEW_CALC_VALUE,
	 PARAM_USED=V_NEW_DYNA_PARAM_VALUE,
	 PREVIOUS_VALUE=V_OLD_CALC_VALUE,
	 PREVIOUS_PARAM_USED=V_DYNA_PARAM_LASTVAL,
	 EVOLUTION_FLAG=SIGN(V_NEW_CALC_VALUE-V_OLD_CALC_VALUE)
  WHERE KPI_ID=V_KPI_ID
  AND (PARAM_USED IS NULL OR PARAM_USED <>  V_NEW_DYNA_PARAM_VALUE);

  UPDATE KPI_CALCULATED_VALUES
  SET    CALCULATED_DATE=SYSDATE,
	 CALCULATED_VALUE=V_NEW_CALC_VALUE,
	 EVOLUTION_FLAG=SIGN(V_NEW_CALC_VALUE-DECODE(PREVIOUS_VALUE,-1, 0, PREVIOUS_VALUE))
  WHERE KPI_ID=V_KPI_ID
  AND PARAM_USED=V_NEW_DYNA_PARAM_VALUE;

  DELETE KPI_CALCULATED_VALUES_HISTORY
  WHERE KPI_ID=V_KPI_ID
  AND PARAM_USED=V_NEW_DYNA_PARAM_VALUE;

  INSERT INTO KPI_CALCULATED_VALUES_HISTORY
  VALUES (
	V_KPI_ID,
	DECODE(V_DYNA_PARAM_TYPE, 4, TO_DATE(V_NEW_DYNA_PARAM_VALUE, V_DYNA_PARAM_FORMAT), 5, TO_DATE(V_NEW_DYNA_PARAM_VALUE, V_DYNA_PARAM_FORMAT), SYSDATE),
	V_NEW_CALC_VALUE,
	V_NEW_DYNA_PARAM_VALUE
  );

   V_DYNA_PARAM_LASTVAL:=V_NEW_DYNA_PARAM_VALUE;
   V_OLD_CALC_VALUE:=V_NEW_CALC_VALUE;
   V_DYNA_PARAM_LASTVAL:=V_NEW_DYNA_PARAM_VALUE;

  END LOOP;


END;
/

